//-----------------------------------------------------------------------------
// name: bart2.ck
// desc: YESBART II. A sequencer. Also a sequel to A to JFK. (v1.1.1)
// 
// author: Peter Li (https://ccrma.stanford.edu/~ptz/)
//         Andrew Zhu Aday (https://ccrma.stanford.edu/~azaday/)
//   date: 12/21/2024
//-----------------------------------------------------------------------------

// Global sequencer params ====================================================

[120.0, 7.5, 30, 90, 180, 60, 30] @=> float BPM[];  // beats per minute
dur STEP[7];  // step durations
for (int i; i < BPM.size(); i++) {
    (1.0/BPM[i])::minute / 2.0 => STEP[i];
}

[
    -1, 1, 6, 11, 13, 18, 23, 25, 30, 25, 23, 18, 13, 11, 6, 1,
    -5, 2, 6, 11, 14, 18, 23, 26, 30, 26, 23, 18, 14, 11, 6, 2,
    1, 8, 11, 16, 20, 23, 28, 32, 35, 32, 28, 23, 20, 16, 11, 8,
    0, 7, 11, 16, 19, 23, 28, 31, 35, 31, 28, 23, 19, 16, 11, 7
] @=> int SCALE1[];  // this chord progression might sound familiar
[ -1, -5, 1, 0 ] @=> int SCALE2[];  // bass notes
[18, 13, 11, 13] @=> int SCALE3[];
[26, 23, 18, 26, 23, 18, 25, 23, 18, 25, 23, 18] @=> int SCALE4[];
[38, 37, 35] @=> int SCALE5[];
[42, 30, 35, 30, 37, 30, 38, 30] @=> int SCALE6[];

Gain input;
SinOsc red => input => dac;
SawOsc green => input => dac;
TriOsc yellow => input => dac;
SinOsc magenta => input => dac;
TriOsc blue => input => dac;
SinOsc pink => input => dac;
SawOsc silver => input => dac;
[red, green, yellow, magenta, blue, pink, silver] @=> Osc osc[];
[0.025, 0.001, 0.0075, 0.025, 0.005, 0.005, 0] @=> float gains[];
48 => int offset;

// Scene setup ================================================================

GWindow.title( "YESBART II" );
GG.scene() @=> GScene @ scene;
GG.camera() @=> GCamera @ cam;
cam.orthographic();  // Orthographic camera mode for 2D scene
GG.scene().light().rotX(1.2);  // for true black background
GG.fps(120);

// Light dark mode toggle
Color.BLACK => vec3 backgroundColor;
Color.WHITE => vec3 transferColor;
0 => int isLight;  // 0 for dark mode, 1 for light mode

// Global constants
17 => int N_COLUMNS;
13 => int N_ROWS;
7 => int N_COLORS;

// data structures for grid system
GCircle @ dots[N_COLUMNS][N_ROWS];
GPlane @ linesX[N_COLUMNS - 1][N_ROWS];
GPlane @ linesY[N_COLUMNS][N_ROWS - 1];

// set up grid system
fun void placeDots() {
    for (-8 => int i; i < 9; i++) {
        for (-6 => int j; j < 7; j++) {
            // Models stations
            GCircle dot --> GG.scene();
            dot.posX(i * 0.5);
            dot.posY(j * 0.5);
            dot.sca(0.3);
            dot @=> dots[i + 8][j + 6];
            // Models horizontal lines
            if (i < 8) {
                GPlane lineX --> GG.scene();
                lineX.posX(i * 0.5 + 0.25);
                lineX.posY(j * 0.5);
                lineX.sca(0.05);
                lineX.scaX(0.2);
                lineX @=> linesX[i + 8][j + 6];
            }
            // Models vertical lines
            if (j < 6) {
                GPlane lineY --> GG.scene();
                lineY.posX(i * 0.5);
                lineY.posY(j * 0.5 + 0.25);
                lineY.sca(0.05);
                lineY.scaY(0.2);
                lineY @=> linesY[i + 8][j + 6];
            }
        }
    }
    GPlane river1 --> GG.scene();
    river1.posX(0.25);
    river1.posY(-3.1);
    river1.posZ(0.00001);
    river1.sca(0.25);
    river1.scaY(5.5);
    river1.color(Color.SKYBLUE);

    GPlane river2 --> GG.scene();
    river2.posX(-1.75);
    river2.posY(1.5);
    river2.posZ(0.00001);
    river2.rotZ(0.785398163);
    river2.sca(0.2);
    river2.scaY(5.5);
    river2.color(Color.SKYBLUE);

    GPlane river3 --> GG.scene();
    river3.posX(1.25);
    river3.posY(0.5);
    river3.posZ(0.00001);
    river3.rotZ(-0.785398163);
    river3.sca(0.2);
    river3.scaY(2.92);
    river3.color(Color.SKYBLUE);

    GPlane river4 --> GG.scene();
    river4.posX(2.25);
    river4.posY(2.96);
    river4.posZ(0.00001);
    river4.sca(0.2);
    river4.scaY(3);
    river4.color(Color.SKYBLUE);
}
placeDots();

// Visual stuff ================================================================

// Global variables
1 => int whichColor;  // color picker
0 => int soundSwitch;  // determines if sequencers should be sporked or exited
int activeDots[N_COLUMNS][N_ROWS];  // tracks which dots are clicked. If clicked, once or twice
int isUpdated[N_COLUMNS][N_ROWS];  // tracks which dot statuses are recently updated
int litDots[N_COLUMNS][N_ROWS];  // tracks which dots should be temporarily colored
GCircle @ colorsToDots[N_COLORS][0];  // tracks active dots for each color
Shred childShreds[N_COLORS];  // stores sporked sequencers when enabled
// color map
[
    Color.RED,
    Color.GREEN,
    Color.YELLOW,
    Color.MAGENTA,
    Color.BLUE,
    Color.PINK,
    Color.DARKGRAY
] @=> vec3 colorMap[];

// get x coord for given GCircle
fun int getX(GCircle dot) {
    return (2 * dot.posX() + 8) $ int;
}

// get y coord for given GCircle
fun int getY(GCircle dot) {
    return (2 * dot.posY() + 6) $ int;
}

// handle color change for transfer -> active
fun void recolor(int c, GCircle dot, int iEach, int jEach) {
    0 => int colorFound;
    for (int cEach; cEach < colorsToDots.size(); cEach++) {
        if (colorFound) break;
        if (cEach != c) {
            for (int dEach; dEach < colorsToDots[cEach].size(); dEach++) {
                if (colorsToDots[cEach][dEach] == dot) {
                    cEach + 1 => activeDots[iEach][jEach];
                    1 => colorFound;
                    break;
                }
            }
        } 
    }
}

// removes all dots and lines of selected color from grid
fun void clearColor(int c) {
    for (int d; d < colorsToDots[c].size(); d++) {
        colorsToDots[c][d] @=> GCircle dot;
        getX(dot) => int iEach;
        getY(dot) => int jEach;
        if (isTransfer(iEach, jEach)) {
            recolor(c, dot, iEach, jEach);
        } else {
            0 => activeDots[iEach][jEach];
        }
    }
    <<< "Color", c + 1, "cleared" >>>;
}

// switches between black and white background
fun void toggleLightDark() {
    if (isLight) {
        GG.scene().light().rotX(1.2);
        Color.BLACK => backgroundColor;
        Color.WHITE => transferColor;
        GG.scene().backgroundColor(backgroundColor * 4);
        0 => isLight;
        <<< "Switched to dark mode" >>>;
    } else {
        GG.scene().light().rotX(0);
        Color.WHITE => backgroundColor;
        Color.BLACK => transferColor;
        GG.scene().backgroundColor(backgroundColor * 4);
        1 => isLight;
        <<< "Switched to light mode" >>>;
    }
}

// switches sequencers on/off
fun void play() {
    if (!soundSwitch) {
        for (int i; i < colorsToDots.size(); i++) {
            spork ~ sequenceBeat(i) @=> childShreds[i];
        }
        1 => soundSwitch;
        <<< "sequencer on" >>>;
    } else {
        for (int i; i < colorsToDots.size(); i++) {
            0 => osc[i].gain;
            childShreds[i].exit();
        }
        0 => soundSwitch;
        <<< "sequencer off" >>>;
    }
}

// change tempo of all sequencers
fun void updateBPM(int op) {
    for (int i; i < BPM.size(); i++) {
        if (op) {
            BPM[i] / 500 +=> BPM[i];
        } else if (BPM[i] > 10) {
            BPM[i] / 500 -=> BPM[i];
        }
        (1.0/BPM[i])::minute / 2.0 => STEP[i];
    }
}

// keyboard controls
fun void kbListener() {
    if (GWindow.keyDown(GWindow.Key_1)) {
        1 => whichColor;
    } else if (GWindow.keyDown(GWindow.Key_2)) {
        2 => whichColor;
    } else if (GWindow.keyDown(GWindow.Key_3)) {
        3 => whichColor;
    } else if (GWindow.keyDown(GWindow.Key_4)) {
        4 => whichColor;
    } else if (GWindow.keyDown(GWindow.Key_5)) {
        5 => whichColor;
    } else if (GWindow.keyDown(GWindow.Key_6)) {
        6 => whichColor;
    } else if (GWindow.keyDown(GWindow.Key_7)) {
        7 => whichColor;
    } else if (GWindow.key(GWindow.Key_Up)) {
        updateBPM(1);
    } else if (GWindow.key(GWindow.Key_Down)) {
        updateBPM(0);
    } else if (GWindow.keyDown(GWindow.Key_Left)) {
        offset--;
    } else if (GWindow.keyDown(GWindow.Key_Right)) {
        offset++;
    } else if (GWindow.keyDown(GWindow.Key_Backspace)) {
        clearColor(whichColor - 1);
    } else if (GWindow.keyDown(GWindow.Key_Enter)) {
        play();
    } else if (GWindow.keyDown(GWindow.Key_Space)) {
        toggleLightDark();
    }
}

// Checks if a dot is colored
fun int isActive(int i, int j) {
    return activeDots[i][j] > 0;
}

// Checks if a dot is a transfer (intersection of 2 or more lines)
fun int isTransfer(int i, int j) {
    return activeDots[i][j] == 999;
}

// Checks if 2 dots have different non black/white colors
fun int isDiffActiveColors(int c1, int c2) {
    return c1 != c2 && c1 > 0 && c2 > 0 && c1 != 999 && c2 != 999;
}

// Checks if a dot should be colored white
fun int shouldBeTransfer(int i, int j) {
    int neighbors[0];
    if (i > 0) {
        neighbors << activeDots[i - 1][j];
    }
    if (i < N_COLUMNS - 1) {
        neighbors << activeDots[i + 1][j];
    }
    if (j > 0) {
        neighbors << activeDots[i][j - 1];
    }
    if (j < N_ROWS - 1) {
        neighbors << activeDots[i][j + 1];
    }
    for (int i; i < neighbors.size(); i++) {
        for (i + 1 => int j; j < neighbors.size(); j++) {
            if (isDiffActiveColors(neighbors[i], neighbors[j])) {
                return 1;
            }
        }
    }
    return 0;
}

// Flashes all dots on the line where a train goes through a transfer
fun void flashTransfer(int c) {
    for (int i; i < colorsToDots[c].size(); i++) {
        colorsToDots[c][i] @=> GCircle dot;
        getX(dot) => int iEach;
        getY(dot) => int jEach;
        2 => litDots[iEach][jEach];
    }
}

// Make transfer dots bigger
fun void resizeTransfer(int i, int j) {
    dots[i][j] @=> GCircle dot;
    dot.sca(0.4);
    dot.posZ(0.00003);
}

// Make dots regular size once they're no longer transfers
fun void resizeInactive(int i, int j) {
    dots[i][j] @=> GCircle dot;
    dot.sca(0.3);
    dot.posZ(0);
}

// toggle between inactive, active, and transfer for each dot
fun void updateStatus(int i, int j) {
    if (!isActive(i, j)) {
        whichColor => activeDots[i][j];  // mark dot as active
    } else if (!isTransfer(i, j) && shouldBeTransfer(i, j)) {
        999 => activeDots[i][j];  // mark dot as transfer
    } else {
        0 => activeDots[i][j];  // mark dot as inactive
    }
}

// colors active dots and surrounding lines according to status of neighbors
fun void colorActive(int i, int j, vec3 color) {
    dots[i][j] @=> GCircle dot;
    dot.color(color);
    if (i < N_COLUMNS - 1) {  // color horizontal lines
        linesX[i][j] @=> GPlane lineX;
        if (isTransfer(i, j) && isActive(i + 1, j)) {
            // transfer stop - color line with neighbor dot color instead of white
            lineX.color(dots[i + 1][j].color());
            lineX.posZ(0.00002);
            resizeTransfer(i, j);
        } else if (isTransfer(i + 1, j) || activeDots[i + 1][j] == activeDots[i][j]) {
            // color line when dot colors match or one or both dots is/are transfer(s)
            lineX.color(color);
            lineX.posZ(0.00002);
            dot.posZ(0.00002);
        } else {
            // erase line when dot becomes inactive
            lineX.color(backgroundColor * 5.5);
            lineX.posZ(0);
        }
    }
    if (j < N_ROWS - 1) {  // color vertical lines
        linesY[i][j] @=> GPlane lineY;
        if (isTransfer(i, j) && isActive(i, j + 1)) {
            // transfer stop - color line with neighbor dot color instead of white
            lineY.color(dots[i][j + 1].color());
            lineY.posZ(0.00002);
            resizeTransfer(i, j);
        } else if (isTransfer(i, j + 1) || activeDots[i][j + 1] == activeDots[i][j]) {
            // color line when dot colors match or one or both dots is/are transfer(s)
            lineY.color(color);
            lineY.posZ(0.00002);
            dot.posZ(0.00002);
        } else {
            // erase line when dot becomes inactive
            lineY.color(backgroundColor * 5.5);
            lineY.posZ(0);
        }
    }
}

// returns the index of dot in array if found, -1 if not found
fun int findDot(GCircle colorToDots[], GCircle dot) {
    for (int j; j < colorToDots.size(); j++) {
        if (dot == colorToDots[j]) {
            return j;
        }
    }
    return -1;
}

// handler for dot coloring
fun void handleActive(int i, int j) {
    dots[i][j] @=> GCircle dot;
    activeDots[i][j] => int color;
    if (isTransfer(i, j)) {
        colorActive(i, j, transferColor);
        whichColor => color;
    } else {
        colorActive(i, j, colorMap[color - 1]);
    }
    if (findDot(colorsToDots[color - 1], dot) == -1) {
        colorsToDots[color - 1] << dot;
    }
}

// colors non-clicked dots and its lines black
fun void handleInactive(int i, int j) {
    dots[i][j] @=> GCircle dot;
    dot.color(backgroundColor * 5.5);
    resizeInactive(i, j);
    if (i < N_COLUMNS - 1) {
        linesX[i][j] @=> GPlane lineX;
        lineX.color(backgroundColor * 5.5);
        lineX.posZ(0);
    }
    if (j < N_ROWS - 1) {
        linesY[i][j] @=> GPlane lineY;
        lineY.color(backgroundColor * 5.5);
        lineY.posZ(0);
    }
    0 => litDots[i][j];
}

// detects when mouse is hovering over a dot and checks for mouse click
fun void handleHover(vec3 mouse_world_pos, int i, int j) {
    dots[i][j] @=> GCircle dot;
    if (mouse_world_pos.x < dot.posX() + dot.scaX() / 2.0 && mouse_world_pos.x > dot.posX() - dot.scaX() / 2.0
        &&
        mouse_world_pos.y < dot.posY() + dot.scaY() / 2.0 && mouse_world_pos.y > dot.posY() - dot.scaY() / 2.0) {
        UI.setMouseCursor(UI_MouseCursor.Hand);
        if (GWindow.mouseLeft()) {
            if (!isUpdated[i][j]) {
                updateStatus(i, j);
                1 => isUpdated[i][j];
            }
        } else {
            0 => isUpdated[i][j];
        }
        if (!isActive(i, j)) {
            dot.color(colorMap[whichColor - 1]);
        }
    }
}

// detects what dots should be temporarily colored pink/violet
fun void handleLit(int i, int j) {
    dots[i][j] @=> GCircle dot;
    litDots[i][j] => int litStatus;
    if (litStatus > 0) {
        resizeTransfer(i, j);
        if (litStatus == 1) {
            dot.color(Color.PINK); 
        } else if (litStatus == 2) {
            dot.color(Color.VIOLET);
        } 
    } else if (!isTransfer(i, j)) {
        dot.sca(0.3);
    }
}

// main handler that sets all colors
fun void setGrid(vec3 mouse_world_pos) {
    for (int i; i < N_COLUMNS; i++) {
        for (int j; j < N_ROWS; j++) {
            if (isActive(i, j)) {  // dot is active/transfer
                handleActive(i, j);
            } else {  // dot is inactive
                handleInactive(i, j);
            }
            handleHover(mouse_world_pos, i, j);  // detect hover
            handleLit(i, j);
        }
    }
}

// Sequencer stuff ===========================================================

SndBuf open("wavs/doorOpen.wav") => input => dac;
SndBuf close("wavs/doorClose.wav") => input => dac;
SndBuf transfer("wavs/transfer.wav") => input => dac;
SndBuf transfer2("wavs/transfer2.wav") => input => dac;
SndBuf transfer3("wavs/transfer3.wav") => input => dac;
0 => open.rate;
0.025 => open.gain;
0 => close.rate;
0.025 => close.gain;
0 => transfer.rate;
0.0125 => transfer.gain;
0 => transfer2.rate;
0.0125 => transfer2.gain;
0 => transfer3.rate;
0.025 => transfer3.gain;

// returns number of active dots for given color
fun int numDotsPerColor(int c) {
    return colorsToDots[c].size();
}

// checks if an active dot has been marked inactive since the last frame and removes it if so
fun void resetRoutes() {
    for (int c; c < colorsToDots.size(); c++) {
        for (numDotsPerColor(c) - 1 => int d; d >= 0; d--) {
            getX(colorsToDots[c][d]) => int i;
            getY(colorsToDots[c][d]) => int j;
            if (!isActive(i, j)) {
                colorsToDots[c].erase(d);
            }
        }
    }
}

// set initial gains
for (int i; i < osc.size(); i++) {
    0 => osc[i].gain;
}

// Function to sort the array by x and then by y as needed
fun void sortActiveDots(GCircle dotsPerColor[]) {
    // Sort the array by x in increasing order
    for (0 => int i; i < dotsPerColor.size() - 1; i++) {
        for (i + 1 => int j; j < dotsPerColor.size(); j++) {
            if (dotsPerColor[i].posX() > dotsPerColor[j].posX() || 
                (dotsPerColor[i].posX() == dotsPerColor[j].posX() && dotsPerColor[i].posY() > dotsPerColor[j].posY())) {
                dotsPerColor[i] @=> GCircle temp;
                dotsPerColor[j] @=> dotsPerColor[i];
                temp @=> dotsPerColor[j];
            }
        }
    }

    // Traverse through the sorted array and sort groups with the same x by y
    for (0 => int i; i < dotsPerColor.size(); ) {
        i => int startIndex;
        dotsPerColor[i].posX() => float currentX;

        // Find the range of GCircles with the same x value
        while (i < dotsPerColor.size() && dotsPerColor[i].posX() == currentX) {
            i++;
        }
        i - 1 => int endIndex;

        1 => int sortOrder; // determines sort direction for y. 1 for increasing, -1 for decreasing
        dotsPerColor[startIndex - 1].posY() => float prevY;

        if (startIndex == 0 && i < dotsPerColor.size()) { // first group
            dotsPerColor[i].posY() => float nextY;
            if (dotsPerColor[i - 1].posY() > nextY) {
                -1 => sortOrder;
            } else {
                1 => sortOrder;
            }
        } else if (i == dotsPerColor.size()) { // last group
            if (startIndex > 0 && prevY > dotsPerColor[startIndex].posY()) {
                -1 => sortOrder;
            }
        } else if (startIndex > 0) {
            dotsPerColor[i].posY() => float nextY;
            if (prevY > nextY) {
                -1 => sortOrder;
            } else {
                1 => sortOrder;
            }
        }

        // Sort the group with the same x value by y in the determined order
        for (startIndex => int m; m < endIndex; m++) {
            for (m + 1 => int n; n <= endIndex; n++) {
                if ((sortOrder == 1 && dotsPerColor[m].posY() > dotsPerColor[n].posY()) || 
                    (sortOrder == -1 && dotsPerColor[m].posY() < dotsPerColor[n].posY())) {
                    dotsPerColor[m] @=> GCircle temp;
                    dotsPerColor[n] @=> dotsPerColor[m];
                    temp @=> dotsPerColor[n];
                }
            }
        }
    }
}

// this will do before I find a better way to write this...
fun int[] getScale(int c) {
    if (c == 0) {
        return SCALE1;
    } else if (c == 1) {
        return SCALE2;
    } else if (c == 2) {
        return SCALE3;
    } else if (c == 3) {
        return SCALE4;
    } else if (c == 4) {
        return SCALE5;
    } else if (c == 5) {
        return SCALE6;
    } else {
        return [-1];
    }
}

// play 1 of 4 audio clips at random 10% of the time
fun void playSamples() {
    Math.random2f(0, 1) => float p;
    if (p > 0.9) {
        Math.random2f(0, 1) => float p2;
        if (p2 < 0.25) {
            if (transfer.pos() == 0 || transfer.pos() == transfer.samples()) {
                1 => transfer.rate;
                0 => transfer.pos;
            }
        } else if (p2 < 0.5) {
            if (transfer2.pos() == 0 || transfer2.pos() == transfer2.samples()) {
                1 => transfer2.rate;
                0 => transfer2.pos;
            }
        } else if (p2 < 0.75) {
            if (transfer3.pos() == 0 || transfer3.pos() == transfer3.samples()) {
                1 => transfer3.rate;
                0 => transfer3.pos;
            }
        } else {
            if (close.pos() == 0 || close.pos() == close.samples()) {
                1 => close.rate;
                0 => close.pos;
            }
        }
    }
}

// spork this to make sound!
fun void sequenceBeat(int c) {
    while (true) {
        if (numDotsPerColor(c) > 0) {
            sortActiveDots(colorsToDots[c]);
            gains[c] => osc[c].gain;
            0 => int note;
            0 => int d;
            0 => int dotDecreasing;  // start out as increasing
            0 => int drumBeat;
            while (numDotsPerColor(c) > 0) {
                for (int dotIndex; dotIndex < colorsToDots[c].size(); dotIndex++) {
                    getX(colorsToDots[c][dotIndex]) => int iEach;
                    getY(colorsToDots[c][dotIndex]) => int jEach;
                    if (dotIndex != d) {
                        0 => litDots[iEach][jEach];
                    }
                }
                getX(colorsToDots[c][d]) => int i;
                getY(colorsToDots[c][d]) => int j;
                <<< "[Color", c + 1 + "] [BPM", BPM[c] + "] playing row", j + 1, "column", i + 1 >>>;
                dots[i][j] @=> GCircle dot;
                1 => litDots[i][j];
                if (isTransfer(i, j)) {
                    flashTransfer(c);
                    playSamples();
                }
                // play sound
                if (c == 6) {
                    if (drumBeat == 3) {
                        1 => open.rate;
                        0 => open.pos;
                        0 => drumBeat;
                    } else {
                        drumBeat++;
                    }  
                } else {
                    getScale(c) @=> int SCALE[];
                    SCALE.size() => int numNotes;
                    if (note == numNotes) {
                        0 => note;
                    }
                    SCALE[note] + offset => float freq;
                    Std.mtof(freq) => osc[c].freq;
                    note++;
                }
                // update dot parser
                if (d == 0) {
                    0 => dotDecreasing;
                }
                if (numDotsPerColor(c) > 1 && d == numDotsPerColor(c) - 1) {
                    1 => dotDecreasing;
                }
                if (dotDecreasing) {
                    d--;
                } else if (d < numDotsPerColor(c) - 1) {
                    d++;
                }
                STEP[c] => now;
            }
        } else {
            0 => osc[c].gain;
            STEP[c] => now;
        }
    }
}

// Game loop ==================================================================

while (true)
{
    GG.nextFrame() => now;

    // get the mouse world pos
    GG.scene().camera().screenCoordToWorldPos(GWindow.mousePos(), 1) => vec3 mouse_world_pos;

    // set state of each GCircle and GPlane
    setGrid(mouse_world_pos);

    // reset array of active dots for sequencer
    resetRoutes();

    // control color picker with number keys (1-7)
    kbListener();

    // uncomment if you want to unravel the scenegraph to see all the GGens
    // if (UI.begin("")) { UI.scenegraph(scene); } UI.end();
}
